@@ -13,8 +13,10 @@
 #include <string.h>
 #include <strings.h>
 #include <errno.h>
+#include <fnmatch.h>
 #include <stdarg.h>
 #include <stdlib.h>
+#include <dirent.h>
 #include <ctype.h>
 #include <unistd.h>
 #include <getopt.h>
@@ -69,6 +71,7 @@
 #define CFG_RECV_BUFSIZE "recv-bufsize"
 #define CFG_SEND_BUFSIZE "send-bufsize"
 #define CFG_LOG_FILENAME "log-filename"
+#define CFG_LOG_LEVEL "log-level"
 #define CFG_RING_SLOTS "ring-slots"
 #define CFG_RING_DATA_LEN "ring-data-len"
 #define CFG_PIDFILE "pidfile"
@@ -180,6 +183,7 @@ config_new(void)
 	r->WRITE_IP_OCTET     = 0;
 	r->WRITE_PROXY_LINE_V1= 0;
 	r->WRITE_PROXY_LINE_V2= 0;
+	r->PROXY_TLV          = 1;
 	r->PROXY_PROXY_LINE   = 0;
 	r->ALPN_PROTOS        = NULL;
 	r->ALPN_PROTOS_LV     = NULL;
@@ -197,6 +201,7 @@ config_new(void)
 	r->CERT_DEFAULT	      = NULL;
 	r->CERT_FILES         = NULL;
 	r->LISTEN_ARGS        = NULL;
+	r->PEM_DIR            = NULL;
 	fa = front_arg_new();
 	fa->port = strdup("8443");
 	fa->pspec = strdup("default");
@@ -220,7 +225,7 @@ config_new(void)
 	r->SHCUPD_MCASTTTL    = NULL;
 #endif
 
-	r->QUIET              = 0;
+	r->LOG_LEVEL          = 0;
 	r->SYSLOG             = 0;
 	r->SYSLOG_FACILITY    = LOG_DAEMON;
 	r->TCP_KEEPALIVE_TIME = 3600;
@@ -277,7 +282,8 @@ config_destroy(hitch_config *cfg)
 	free(cfg->OCSP_DIR);
 	free(cfg->ALPN_PROTOS);
 	free(cfg->ALPN_PROTOS_LV);
-
+	free(cfg->PEM_DIR);
+	free(cfg->PEM_DIR_GLOB);
 #ifdef USE_SHARED_CACHE
 	int i;
 	free(cfg->SHCUPD_IP);
@@ -322,26 +328,58 @@ config_param_val_bool(char *val, int *res)
 	return (1);
 }
 
+static int
+config_param_uds(const char *str, char **path)
+{
+	struct stat st;
+
+	AN(path);
+
+	if (strlen(str) > 104) {
+		config_error_set("UNIX domain socket path too long.");
+		return (0);
+	}
+
+	if (stat(str, &st)) {
+		config_error_set("Unable to stat path '%s': %s", str,
+		    strerror(errno));
+		return (0);
+	}
+
+	if (!S_ISSOCK(st.st_mode)) {
+		config_error_set("Invalid path '%s': Not a socket.", str);
+		return (0);
+	}
+
+	*path = strdup(str);
+	return (1);
+}
+
 static int
 config_param_host_port_wildcard(const char *str, char **addr,
-    char **port, char **cert, int wildcard_okay)
+    char **port, char **cert, int wildcard_okay, char **path)
 {
 	const char *cert_ptr = NULL;
+	char port_buf[PORT_LEN];
+	char addr_buf[ADDR_LEN];
 
 	if (str == NULL) {
 		config_error_set("Invalid/unset host/port string.");
 		return (0);
 	}
 
+	/* UDS addresses start with a '/' */
+	if (path != NULL && *str == '/') {
+		*addr = NULL;
+		*port = NULL;
+		return (config_param_uds(str, path));
+	}
+
 	if (strlen(str) > ADDR_LEN) {
 		config_error_set("Host address too long.");
 		return (0);
 	}
 
-	// address/port buffers
-	char port_buf[PORT_LEN];
-	char addr_buf[ADDR_LEN];
-
 	memset(port_buf, '\0', sizeof(port_buf));
 	memset(addr_buf, '\0', sizeof(addr_buf));
 
@@ -414,9 +452,10 @@ config_param_host_port_wildcard(const char *str, char **addr,
 }
 
 static int
-config_param_host_port(char *str, char **addr, char **port)
+config_param_host_port(char *str, char **addr, char **port, char **path)
 {
-	return (config_param_host_port_wildcard(str, addr, port, NULL, 0));
+	return (config_param_host_port_wildcard(str, addr, port,
+		NULL, 0, path));
 }
 
 
@@ -668,7 +707,7 @@ config_param_shcupd_peer(char *str, hitch_config *cfg)
 	memset(port, '\0', PORT_LEN);
 
 	// try to parse address
-	if (! config_param_host_port(str, &addr, &port)) {
+	if (! config_param_host_port(str, &addr, &port, NULL)) {
 		r = 0;
 		goto outta_parse_peer;
 	}
@@ -789,7 +828,7 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 
 		fa = front_arg_new();
 		r = config_param_host_port_wildcard(v,
-		    &fa->ip, &fa->port, &certfile, 1);
+		    &fa->ip, &fa->port, &certfile, 1, NULL);
 		if (r != 0) {
 			if (certfile != NULL) {
 				cert = cfg_cert_file_new();
@@ -809,7 +848,9 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 	} else if (strcmp(k, CFG_BACKEND) == 0) {
 		free(cfg->BACK_PORT);
 		free(cfg->BACK_IP);
-		r = config_param_host_port(v, &cfg->BACK_IP, &cfg->BACK_PORT);
+		free(cfg->BACK_PATH);
+		r = config_param_host_port(v, &cfg->BACK_IP, &cfg->BACK_PORT,
+		    &cfg->BACK_PATH);
 	} else if (strcmp(k, CFG_WORKERS) == 0) {
 		r = config_param_val_long(v, &cfg->NCORES, 1);
 	} else if (strcmp(k, CFG_BACKLOG) == 0) {
@@ -825,7 +866,7 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 	} else if (strcmp(k, CFG_SHARED_CACHE_LISTEN) == 0) {
 		if (strlen(v) > 0)
 			r = config_param_host_port_wildcard(v, &cfg->SHCUPD_IP,
-			    &cfg->SHCUPD_PORT, NULL, 1);
+			    &cfg->SHCUPD_PORT, NULL, 1, NULL);
 	} else if (strcmp(k, CFG_SHARED_CACHE_PEER) == 0) {
 		r = config_param_shcupd_peer(v, cfg);
 	} else if (strcmp(k, CFG_SHARED_CACHE_MCASTIF) == 0) {
@@ -874,7 +915,14 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 			}
 		}
 	} else if (strcmp(k, CFG_QUIET) == 0) {
-		r = config_param_val_bool(v, &cfg->QUIET);
+		int b;
+		r = config_param_val_bool(v, &b);
+		if (b)
+			cfg->LOG_LEVEL = 0;
+		else
+			cfg->LOG_LEVEL = 1;
+	} else if (strcmp(k, CFG_LOG_LEVEL) == 0) {
+		r = config_param_val_int(v, &cfg->LOG_LEVEL, 1);
 	} else if (strcmp(k, CFG_LOG_FILENAME) == 0) {
 		if (strlen(v) > 0) {
 			config_assign_str(&cfg->LOG_FILENAME, v);
@@ -932,10 +980,6 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		r = config_param_val_int(v, &cfg->RECV_BUFSIZE, 1);
 	} else if (strcmp(k, CFG_SEND_BUFSIZE) == 0) {
 		r = config_param_val_int(v, &cfg->SEND_BUFSIZE, 1);
-	} else if (strcmp(k, CFG_LOG_FILENAME) == 0) {
-		if (strlen(v) > 0) {
-			config_assign_str(&cfg->LOG_FILENAME, v);
-		}
 	} else if (strcmp(k, CFG_PIDFILE) == 0) {
 		if (strlen(v) > 0) {
 			config_assign_str(&cfg->PIDFILE, v);
@@ -1073,6 +1117,53 @@ config_disp_log_facility (int facility)
 	}
 }
 
+int
+config_scan_pem_dir(char *pemdir, hitch_config *cfg)
+{
+	int n, i;
+	struct dirent **d;
+
+	n = scandir(pemdir, &d, NULL, alphasort);
+	if (n < 0) {
+		config_error_set("Unable to open directory '%s': %s", pemdir,
+		    strerror(errno));
+		return (1);
+	}
+	for (i = 0; i < n; i++) {
+		struct cfg_cert_file *cert;
+		char fpath[PATH_MAX + NAME_MAX];
+
+		if (cfg->PEM_DIR_GLOB != NULL) {
+			if (fnmatch(cfg->PEM_DIR_GLOB, d[i]->d_name, 0))
+				continue;
+		}
+		if (d[i]->d_type != DT_REG)
+			continue;
+
+		strncpy(fpath, pemdir, PATH_MAX + NAME_MAX - 1);
+		strncat(fpath, d[i]->d_name,
+		    PATH_MAX + NAME_MAX - strlen(fpath) - 1);
+
+		cert = cfg_cert_file_new();
+		config_assign_str(&cert->filename, fpath);
+		int r = cfg_cert_vfy(cert);
+		if (r != 0) {
+			/* If no default has been set, use the first
+			 * match according to alphasort  */
+			if (cfg->CERT_DEFAULT == NULL)
+				cfg->CERT_DEFAULT = cert;
+			else
+				cfg_cert_add(cert, &cfg->CERT_FILES);
+		} else {
+			cfg_cert_file_free(&cert);
+		}
+		free(d[i]);
+	}
+
+	free(d);
+	return (0);
+}
+
 void
 config_print_usage_fd(char *prog, FILE *out)
 {
@@ -1101,6 +1192,8 @@ config_print_usage_fd(char *prog, FILE *out)
 	fprintf(out, "\n");
 	fprintf(out, "      --client                Enable client proxy mode\n");
 	fprintf(out, "  -b  --backend=[HOST]:PORT   Backend [connect] (default is \"%s\")\n", config_disp_hostport(cfg->BACK_IP, cfg->BACK_PORT));
+	fprintf(out, "                              The -b argument can also take a UNIX domain socket path\n");
+	fprintf(out, "                              E.g. --backend=\"/path/to/sock\"\n");
 	fprintf(out, "  -f  --frontend=[HOST]:PORT[+CERT]    Frontend [bind] (default is \"%s\")\n", config_disp_hostport(cfg->LISTEN_DEFAULT->ip, cfg->LISTEN_DEFAULT->port));
 	fprintf(out, "                                (Note: brackets are mandatory in endpoint specifiers.)\n");
 	fprintf(out, "      --recv-bufsize=SIZE    Receive buffer size on client socket (Default: %d)\n", cfg->RECV_BUFSIZE);
@@ -1140,7 +1233,9 @@ config_print_usage_fd(char *prog, FILE *out)
 	fprintf(out, "  -g  --group=GROUP          Set gid after binding the socket (Default: \"%s\")\n", config_disp_gid(cfg->GID));
 	fprintf(out, "\n");
 	fprintf(out, "LOGGING:\n");
-	fprintf(out, "  -q  --quiet                Be quiet; emit only error messages\n");
+	fprintf(out, "  -q  --quiet                Be quiet; emit only error messages (deprecated, use 'log-level')\n");
+	fprintf(out, "  -L  --log-level=NUM        Log level. 0=silence, 1=err, 2=info/debug (Default: %d)\n",
+		cfg->LOG_LEVEL);
 	fprintf(out, "  -l  --log-filename=FILE    Send log message to a logfile instead of stderr/stdout\n");
 	fprintf(out, "  -s  --syslog               Send log message to syslog in addition to stderr/stdout\n");
 	fprintf(out, "      --syslog-facility=FACILITY    Syslog facility to use (Default: \"%s\")\n", config_disp_log_facility(cfg->SYSLOG_FACILITY));
@@ -1236,7 +1331,7 @@ create_alpn_callback_data(hitch_config *cfg, char **error)
 }
 
 int
-config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
+config_parse_cli(int argc, char **argv, hitch_config *cfg)
 {
 	static int tls = 0, ssl = 0;
 	static int client = 0;
@@ -1244,9 +1339,6 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 
 	optind = 1;
 
-	AN(retval);
-	*retval = 0;
-
 	struct option long_options[] = {
 		{ CFG_CONFIG, 1, NULL, CFG_PARAM_CFGFILE },
 		{ "tls", 0, &tls, 1},
@@ -1272,6 +1364,7 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		{ CFG_GROUP, 1, NULL, 'g' },
 		{ CFG_QUIET, 0, NULL, 'q' },
 		{ CFG_LOG_FILENAME, 1, NULL, 'l' },
+		{ CFG_LOG_LEVEL, 1, NULL, 'L' },
 		{ CFG_SYSLOG, 0, NULL, 's' },
 		{ CFG_SYSLOG_FACILITY, 1, NULL, CFG_PARAM_SYSLOG_FACILITY },
 		{ CFG_SEND_BUFSIZE, 1, NULL, CFG_PARAM_SEND_BUFSIZE },
@@ -1290,11 +1383,10 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		{ "help", 0, NULL, 'h' },
 		{ 0, 0, 0, 0 }
 	};
-#define SHORT_OPTS "c:e:Ob:f:n:B:l:C:U:p:P:M:k:r:u:g:qstVho:R:"
+#define SHORT_OPTS "c:e:Ob:f:n:B:l:L:C:U:p:P:M:k:r:u:g:qstVho:R:"
 
 	if (argc == 1) {
 		config_print_usage(argv[0]);
-		*retval = 0;
 		return (1);
 	}
 
@@ -1305,18 +1397,17 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		int option_index = 0;
 		c = getopt_long(argc, argv, SHORT_OPTS,
 			long_options, &option_index);
-		if (c == -1)
+		if (c == -1) {
 			break;
+		}
 		else if (c == '?') {
 			config_error_set("Invalid command line parameters. "
 			    "Run %s --help for instructions.",
 			    basename(argv[0]));
-			*retval = 1;
 			return (1);
 		}
 		else if (c == CFG_PARAM_CFGFILE) {
 			if (config_file_parse(optarg, cfg) != 0) {
-				*retval = 1;
 				return (1);
 			}
 		}
@@ -1375,6 +1466,7 @@ CFG_ARG('o', CFG_OCSP_DIR);
 CFG_ON('O', CFG_PREFER_SERVER_CIPHERS);
 CFG_ON('q', CFG_QUIET);
 CFG_ARG('l', CFG_LOG_FILENAME);
+CFG_ARG('L', CFG_LOG_LEVEL);
 CFG_ON('s', CFG_SYSLOG);
 #undef CFG_ARG
 #undef CFG_ON
@@ -1383,12 +1475,10 @@ CFG_ON('s', CFG_SYSLOG);
 			break;
 		case 'V':
 			printf("%s %s\n", basename(argv[0]), VERSION);
-			*retval = 0;
 			return (1);
 			break;
 		case 'h':
 			config_print_usage(argv[0]);
-			*retval = 0;
 			return (1);
 			break;
 
@@ -1396,12 +1486,10 @@ CFG_ON('s', CFG_SYSLOG);
 			config_error_set("Invalid command line parameters. "
 			    "Run %s --help for instructions.",
 			    basename(argv[0]));
-			*retval = 1;
 			return (1);
 		}
 
 		if (ret != 0) {
-			*retval = 1;
 			return (1);
 		}
 	}
@@ -1409,13 +1497,11 @@ CFG_ON('s', CFG_SYSLOG);
 	if ((tls || ssl) && tls_protos_config_file != 0) {
 		config_error_set("Deprecated options --tls and --ssl cannot be"
 		    " used to override tls-protos in a config file.");
-		*retval = 1;
 		return (1);
 	}
 	if (tls && ssl) {
 		config_error_set("Options --tls and --ssl are mutually"
 		    " exclusive.");
-		*retval = 1;
 		return (1);
 	} else {
 		if (ssl)
@@ -1434,20 +1520,14 @@ CFG_ON('s', CFG_SYSLOG);
 		config_error_set("Options --write-ip, --write-proxy-proxy,"
 		    " --write-proxy-v1 and --write-proxy-v2 are"
 		    " mutually exclusive.");
-		*retval = 1;
 		return (1);
 	}
 
 
-	if (cfg->DAEMONIZE) {
-		cfg->QUIET = 1;
-	}
-
 #ifdef USE_SHARED_CACHE
 	if (cfg->SHCUPD_IP != NULL && ! cfg->SHARED_CACHE) {
 		config_error_set("Shared cache update listener is defined,"
 		    " but shared cache is disabled.");
-		*retval = 1;
 		return (1);
 	}
 #endif
@@ -1466,7 +1546,6 @@ CFG_ON('s', CFG_SYSLOG);
 				    " \"%s\" is bad. See man page for more"
 				    " info.",
 				    cfg->ALPN_PROTOS);
-			*retval = 1;
 			return (1);
 		}
 		AN(cfg->ALPN_PROTOS_LV);
@@ -1477,7 +1556,6 @@ CFG_ON('s', CFG_SYSLOG);
 			    " more than one protocol while proxy-v2 is "
 			    " not selected. This is a configuration"
 			    " error.");
-			*retval = 1;
 			return (1);
 			/* Note that this test was carried out indepenently of
 			   the availability of ALPN / NPN */
@@ -1502,7 +1580,6 @@ CFG_ON('s', CFG_SYSLOG);
 			    " %lx, which does not have NPN or ALPN support,"
 			    " yet alpn-protos has been set to %s.",
 			    OPENSSL_VERSION_NUMBER, cfg->ALPN_PROTOS);
-			*retval = 1;
 			return (1);
 		}
 		else
@@ -1518,18 +1595,21 @@ CFG_ON('s', CFG_SYSLOG);
 	argv += optind;
 	for (i = 0; i < argc; i++) {
 		if (config_param_validate(CFG_PEM_FILE, argv[i], cfg, NULL, 0)) {
-			*retval = 1;
 			return (1);
 		}
 	}
 
+	if (cfg->PEM_DIR != NULL) {
+		if (config_scan_pem_dir(cfg->PEM_DIR, cfg))
+			return (1);
+	}
+
 	if (cfg->PMODE == SSL_SERVER && cfg->CERT_DEFAULT == NULL) {
 		struct front_arg *fa, *fatmp;
 		HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp)
 			if (HASH_CNT(hh, fa->certs) == 0) {
 				config_error_set("No x509 certificate PEM file "
 				    "specified for frontend '%s'!", fa->pspec);
-				*retval = 1;
 				return (1);
 			}
 	}
