@@ -52,11 +52,13 @@ typedef enum {
 	SSLv3_PROTO	= 0x01,
 	TLSv1_0_PROTO	= 0x02,
 	TLSv1_1_PROTO	= 0x04,
-	TLSv1_2_PROTO	= 0x08
+	TLSv1_2_PROTO	= 0x08,
+	TLSv1_3_PROTO	= 0x10
 } TLS_PROTOCOL;
 
-#define DEFAULT_TLS_PROTOS (TLSv1_1_PROTO | TLSv1_2_PROTO)
-#define TLS_OPTION_PROTOS (TLSv1_0_PROTO | DEFAULT_TLS_PROTOS)
+#define DEFAULT_TLS_PROTOS (TLSv1_2_PROTO | TLSv1_3_PROTO)
+#define TLS_OPTION_PROTOS \
+	(TLSv1_0_PROTO | TLSv1_1_PROTO | DEFAULT_TLS_PROTOS)
 #define SSL_OPTION_PROTOS (SSLv3_PROTO | TLS_OPTION_PROTOS)
 
 typedef enum {
@@ -101,6 +103,7 @@ struct __hitch_config {
     int WRITE_PROXY_LINE_V1;
     int WRITE_PROXY_LINE_V2;
     int PROXY_PROXY_LINE;
+    unsigned PROXY_TLV;
     char *ALPN_PROTOS;
     unsigned char *ALPN_PROTOS_LV;
     unsigned ALPN_PROTOS_LV_LEN;
@@ -111,6 +114,7 @@ struct __hitch_config {
     struct front_arg *LISTEN_DEFAULT;
     char *BACK_IP;
     char *BACK_PORT;
+    char *BACK_PATH;
     long NCORES;
     struct cfg_cert_file *CERT_FILES;
     struct cfg_cert_file *CERT_DEFAULT;
@@ -125,7 +129,7 @@ struct __hitch_config {
     char *SHCUPD_MCASTIF;
     char *SHCUPD_MCASTTTL;
 #endif
-    int QUIET;
+    int LOG_LEVEL;
     int SYSLOG;
     int SYSLOG_FACILITY;
     int TCP_KEEPALIVE_TIME;
@@ -142,6 +146,8 @@ struct __hitch_config {
     char *PIDFILE;
     int SNI_NOMATCH_ABORT;
     int TEST;
+    char *PEM_DIR;
+    char *PEM_DIR_GLOB;
     int OCSP_VFY;
     char *OCSP_DIR;
     double OCSP_RESP_TMO;
@@ -154,6 +160,6 @@ typedef struct __hitch_config hitch_config;
 const char * config_error_get (void);
 hitch_config * config_new (void);
 void config_destroy (hitch_config *cfg);
-int config_parse_cli(int argc, char **argv, hitch_config *cfg, int *rv);
+int config_parse_cli(int argc, char **argv, hitch_config *cfg);
 
 #endif  /* CONFIGURATION_H_INCLUDED */
