@@ -40,6 +40,7 @@
 #include <sys/socket.h>
 #include <sys/time.h>
 #include <sys/types.h>
+#include <sys/un.h>
 #include <sys/wait.h>  /* WAIT_PID */
 
 #ifdef __linux__
@@ -184,6 +185,7 @@ struct frontend {
 	int			sni_nomatch_abort;
 	struct sni_name_s	*sni_names;
 	struct sslctx_s		*ssl_ctxs;
+	struct sslctx_s		*default_ctx;
 	char			*pspec;
 	struct listen_sock_head	socks;
 	VTAILQ_ENTRY(frontend)	list;
@@ -908,6 +910,10 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 		ssloptions |= SSL_OP_NO_TLSv1_1;
 	if (!(selected_protos & TLSv1_2_PROTO))
 		ssloptions |= SSL_OP_NO_TLSv1_2;
+#ifdef SSL_OP_NO_TLSv1_3
+	if (!(selected_protos & TLSv1_3_PROTO))
+		ssloptions |= SSL_OP_NO_TLSv1_3;
+#endif
 	ctx = SSL_CTX_new((CONFIG->PMODE == SSL_CLIENT) ?
 	    SSLv23_client_method() : SSLv23_server_method());
 
@@ -992,7 +998,7 @@ make_ctx_fr(const struct cfg_cert_file *cf, const struct frontend *fr,
 		ERR("Error setting up SNI support.\n");
 	}
 	CHECK_OBJ_ORNULL(fr, FRONTEND_MAGIC);
-	if (!SSL_CTX_set_tlsext_servername_arg(ctx, fr)) {
+	if (!SSL_CTX_set_tlsext_servername_arg(ctx, (void *)fr)) {
 		ERR("Error setting SNI servername arg.\n");
 	}
 
@@ -1208,7 +1214,6 @@ init_certs(void) {
 	sslctx *so;
 
 	if (CONFIG->CERT_DEFAULT != NULL) {
-		AN(CONFIG->CERT_DEFAULT);
 		default_ctx = make_ctx(CONFIG->CERT_DEFAULT);
 		if (default_ctx == NULL)
 			exit(1);
@@ -1225,8 +1230,8 @@ init_certs(void) {
 			so = make_ctx(cf);
 			if (so == NULL)
 				exit(1);
-			HASH_ADD_KEYPTR(hh, ssl_ctxs, cf->filename,
-			    strlen(cf->filename), so);
+			HASH_ADD_KEYPTR(hh, ssl_ctxs, so->filename,
+			    strlen(so->filename), so);
 #ifndef OPENSSL_NO_TLSEXT
 			insert_sni_names(so, &sni_names);
 #endif
@@ -1423,7 +1428,7 @@ create_frontend(const struct front_arg *fa)
 	sslctx *so;
 	int count = 0;
 	struct frontend_head tmp_list;
-	struct cfg_cert_file *cf, *cftmp;
+	struct cfg_cert_file *cf;
 
 	CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
 	ALLOC_OBJ(fr, FRONTEND_MAGIC);
@@ -1441,17 +1446,19 @@ create_frontend(const struct front_arg *fa)
 		return (NULL);
 	}
 
-	HASH_ITER(hh, fa->certs, cf, cftmp) {
+	for (cf = fa->certs; cf != NULL; cf = cf->hh.next) {
 		so = make_ctx_fr(cf, fr, fa);
 		if (so == NULL) {
 			destroy_frontend(fr);
 			return (NULL);
 		}
 		HASH_ADD_KEYPTR(hh, fr->ssl_ctxs,
-		    cf->filename, strlen(cf->filename), so);
+		    so->filename, strlen(so->filename), so);
 #ifndef OPENSSL_NO_TLSEXT
 		insert_sni_names(so, &fr->sni_names);
 #endif
+		if (cf->hh.next == NULL)
+			fr->default_ctx = so;
 	}
 
 	return (fr);
@@ -1469,6 +1476,7 @@ backend_create(struct sockaddr *sa)
 	struct suckaddr *su;
 
 	addr = Get_Sockaddr(sa, &len);
+	AN(addr);
 	su = VSA_Malloc(addr, len);
 	ALLOC_OBJ(b, BACKEND_MAGIC);
 	b->backaddr = su;
@@ -1508,23 +1516,24 @@ static int
 create_back_socket(struct backend *b)
 {
 	socklen_t len;
-	const void *addr;
+	const struct sockaddr *addr;
 
 	CHECK_OBJ_NOTNULL(b, BACKEND_MAGIC);
-	addr = VSA_Get_Sockaddr(b->backaddr, &len);
+	addr = (struct sockaddr *) VSA_Get_Sockaddr(b->backaddr, &len);
 	AN(addr);
-	int s = socket(((const struct sockaddr *)addr)->sa_family,
-		       SOCK_STREAM, IPPROTO_TCP);
+	int s = socket(addr->sa_family, SOCK_STREAM, 0);
 
 	if (s == -1)
 		return -1;
 
-	int flag = 1;
-	int ret = setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
-	    (char *)&flag, sizeof(flag));
-	if (ret == -1)
-		ERR("Couldn't setsockopt to backend (TCP_NODELAY): %s\n",
-		    strerror(errno));
+	if (addr->sa_family != PF_UNIX) {
+		int flag = 1;
+		int ret = setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
+		    (char *)&flag, sizeof(flag));
+		if (ret == -1)
+			ERR("Couldn't setsockopt to backend (TCP_NODELAY):"
+			    " %s\n", strerror(errno));
+	}
 	if (setnonblocking(s) < 0) {
 		(void)close(s);
 		return (-1);
@@ -1716,6 +1725,24 @@ clear_write(struct ev_loop *loop, ev_io *w, int revents)
 
 static void start_handshake(proxystate *ps, int err);
 
+static unsigned
+sockaddr_port(const struct sockaddr *sa)
+{
+	const struct sockaddr_in *sa4;
+	const struct sockaddr_in6 *sa6;
+
+	switch (sa->sa_family) {
+	case PF_INET:
+		sa4 = (struct sockaddr_in *) sa;
+		return (ntohs((sa4->sin_port)));
+	case PF_INET6:
+		sa6 = (struct sockaddr_in6 *) sa;
+		return (ntohs((sa6->sin6_port)));
+	default:
+		return (0);
+	}
+}
+
 /* Continue/complete the asynchronous connect() before starting data
  * transmission between front/backend */
 static void
@@ -1739,15 +1766,15 @@ handle_connect(struct ev_loop *loop, ev_io *w, int revents)
 		ev_timer_stop(loop, &ps->ev_t_connect);
 
 		if (!ps->clear_connected) {
-			struct sockaddr_storage addr;
+			struct sockaddr_storage ss;
 			socklen_t sl;
 
 			sl = sizeof(addr);
 			r = getsockname(ps->fd_down,
-			    (struct sockaddr*) &addr, &sl);
+			    (struct sockaddr *) &ss, &sl);
 			AZ(r);
-			ps->connect_port =
-			    ntohs(((struct sockaddr_in*)&addr)->sin_port);
+			ps->connect_port = sockaddr_port(
+				(struct sockaddr *) &ss);
 			LOGPROXY(ps, "backend connected\n");
 
 			ps->clear_connected = 1;
@@ -1807,7 +1834,7 @@ start_handshake(proxystate *ps, int err)
 }
 
 static void
-get_proto_selected(proxystate *ps, const unsigned char **selected, unsigned *len) {
+get_alpn(proxystate *ps, const unsigned char **selected, unsigned *len) {
 	*selected = NULL;
 	*len = 0;
 #ifdef OPENSSL_WITH_ALPN
@@ -1819,17 +1846,27 @@ get_proto_selected(proxystate *ps, const unsigned char **selected, unsigned *len
 #endif
 }
 
+static int
+proxy_tlv_append(char *dst, ssize_t dstlen, unsigned type,
+    const char *val, ssize_t len)
+{
+	if (len == -1)
+		len = strlen(val);
+	if (dstlen < len + 3)
+		return (0);
+	dst[0] = type;
+	dst[1] = (len >> 8) & 0xff;
+	dst[2] = len & 0xff;
+	memcpy(dst + 3, val, len);
+	return (len + 3);
+}
+
 static void
 write_proxy_v2(proxystate *ps, const struct sockaddr *local)
 {
 	char *base;
-#if defined(OPENSSL_WITH_ALPN) || defined(OPENSSL_WITH_NPN)
-	const unsigned char *selected = NULL;
-	unsigned selected_len = 0;
-	char *alpn_base;
+	int i;
 
-	get_proto_selected(ps, &selected, &selected_len);
-#endif
 	struct ha_proxy_v2_hdr *p;
 	union addr {
 		struct sockaddr		sa;
@@ -1840,19 +1877,14 @@ write_proxy_v2(proxystate *ps, const struct sockaddr *local)
 	CHECK_OBJ_NOTNULL(ps, PROXYSTATE_MAGIC);
 	base = ringbuffer_write_ptr(&ps->ring_ssl2clear);
 	p = (struct ha_proxy_v2_hdr *)base;
-	size_t len = 16;
+	size_t len = 16, maxlen;
 	l = (union addr *) local;
 	r = (union addr *) &ps->remote_ip;
+	maxlen = ps->ring_ssl2clear.data_len;
 
 	memcpy(&p->sig,"\r\n\r\n\0\r\nQUIT\n", 12);
 	p->ver_cmd = 0x21; 	/* v2|PROXY */
 	p->fam = l->sa.sa_family == AF_INET ? 0x11 : 0x21;
-	size_t payload_len = l->sa.sa_family == AF_INET ? 12 : 36;
-#if defined(OPENSSL_WITH_ALPN) || defined(OPENSSL_WITH_NPN)
-	if (selected_len > 0)
-		payload_len += selected_len + 3;
-#endif
-	p->len = htons(payload_len);
 
 	if (l->sa.sa_family == AF_INET) {
 		len += 12;
@@ -1887,17 +1919,56 @@ write_proxy_v2(proxystate *ps, const struct sockaddr *local)
 
 	/* This is where we add something related to NPN or ALPN*/
 #if defined(OPENSSL_WITH_ALPN) || defined(OPENSSL_WITH_NPN)
-	if (selected_len > 0) {
+#define PP2_TYPE_ALPN	0x01
+	const char *alpn_tok = NULL;
+	unsigned alpn_len = 0;
+	get_alpn(ps, (const unsigned char **)&alpn_tok,
+	    &alpn_len);
+	if (alpn_len > 0) {
 		/* let the server know that a protocol was selected. */
-		alpn_base = base + len;
-		alpn_base[0] = 1 /* PP2_TYPE_ALPN */;
-		alpn_base[1] = (selected_len >> 8) & 0xff;
-		alpn_base[2] = selected_len & 0xff;
-		memcpy(alpn_base + 3, selected, selected_len);
-		len += selected_len + 3;
+		i = proxy_tlv_append(base + len, maxlen - len,
+		    PP2_TYPE_ALPN, alpn_tok, alpn_len);
+		AN(i);
+		len += i;
 	}
 #endif
-	assert(len == payload_len + 16);
+	if (CONFIG->PROXY_TLV) {
+#define PP2_TYPE_SSL		0x20
+#define PP2_CLIENT_SSL		0x01
+#define PP2_SUBTYPE_SSL_VERSION	0x21
+#define PP2_SUBTYPE_SSL_CIPHER	0x23
+		char *tlvp = base + len;
+		ssize_t sz = 0;
+		const char *tmp;
+
+		tlvp[0] = PP2_TYPE_SSL;
+		/* tlvp[1..2] to be updated with payload length later */
+		len += 3;
+
+		base[len] = PP2_CLIENT_SSL;
+		base[len + 1] = 1;
+		len += 5;
+		sz += 5;
+
+		tmp = SSL_get_version(ps->ssl);
+		AN(tmp);
+		i = proxy_tlv_append(base + len, maxlen - len,
+		    PP2_SUBTYPE_SSL_VERSION, tmp, -1);
+		len += i;
+		sz += i;
+
+		tmp = SSL_get_cipher_name(ps->ssl);
+		AN(tmp);
+		i = proxy_tlv_append(base + len, maxlen - len,
+		    PP2_SUBTYPE_SSL_CIPHER, tmp, -1);
+		len += i;
+		sz += i;
+
+		tlvp[1] = (sz >> 8) & 0xff;
+		tlvp[2] = sz & 0xff;
+	}
+
+	p->len = htons(len - 16);
 	ringbuffer_write_append(&ps->ring_ssl2clear, len);
 }
 
@@ -1974,26 +2045,26 @@ static int is_protocol_matching(const unsigned char *selected, unsigned len) {
 }
 
 static int is_alpn_shutdown_needed(proxystate *ps) {
-	const unsigned char *selected;
-	unsigned selected_len;
+	const unsigned char *alpn_tok;
+	unsigned alpn_len;
 
 	if (CONFIG->ALPN_PROTOS_LV == NULL)
 		return 0;
 
-	get_proto_selected(ps, &selected, &selected_len);
-	if (selected_len == 0) {
+	get_alpn(ps, &alpn_tok, &alpn_len);
+	if (alpn_len == 0) {
 		/* If alpn / npn was tried, shut down */
 		if(ps->npn_alpn_tried) {
 			LOGPROXY(ps, "Unsuccessful NPN/ALPN negotiation\n");
 			return 1;
 		} else
 			LOGPROXY(ps, "No NPN/ALPN negotiation happened.\n");
-	} else if (!is_protocol_matching(selected, selected_len)) {
+	} else if (!is_protocol_matching(alpn_tok, alpn_len)) {
 		LOGPROXY(ps, "NPN: Unknown protocol selected\n");
 		return 1;
 	} else
 		LOGPROXY(ps, "NPN/ALPN protocol: %.*s\n",
-		    selected_len, selected);
+		    alpn_len, alpn_tok);
 	return 0;
 }
 #endif
@@ -2403,8 +2474,8 @@ handle_accept(struct ev_loop *loop, ev_io *w, int revents)
 
 
 	CAST_OBJ_NOTNULL(fr, w->data, FRONTEND_MAGIC);
-	if (fr->ssl_ctxs != NULL)
-		CAST_OBJ_NOTNULL(so, fr->ssl_ctxs, SSLCTX_MAGIC);
+	if (fr->default_ctx != NULL)
+		CAST_OBJ_NOTNULL(so, fr->default_ctx, SSLCTX_MAGIC);
 	else
 		CAST_OBJ_NOTNULL(so, default_ctx, SSLCTX_MAGIC);
 
@@ -2518,6 +2589,9 @@ Get_Sockaddr(const struct sockaddr *sa, socklen_t *sl)
 	case PF_INET6:
 		*sl = sizeof(struct sockaddr_in6);
 		break;
+	case PF_UNIX:
+		*sl = sizeof(struct sockaddr_un);
+		break;
 	default:
 		*sl = 0;
 		return (NULL);
@@ -2538,7 +2612,7 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 	if (r  == -1) {
 		if (errno == EWOULDBLOCK || errno == EAGAIN)
 			return;
-		SOCKERR("Error in mgt->worker read operation. "
+		LOGL("Error in mgt->worker read operation. "
 		    "Restarting process.");
 		/* If something went wrong here, the process will be
 		 * left in utter limbo as to whether it should keep
@@ -2552,6 +2626,8 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 
 	if (wu.type == WORKER_GEN && wu.payload.gen != worker_gen) {
 		/* This means this process has reached its retirement age. */
+		if (worker_state == WORKER_EXITING)
+			return;
 		worker_state = WORKER_EXITING;
 
 		/* Stop accepting new connections. */
@@ -2568,8 +2644,9 @@ handle_mgt_rd(struct ev_loop *loop, ev_io *w, int revents)
 
 		LOGL("Worker %d (gen: %d): State %s\n", core_id, worker_gen,
 		(worker_state == WORKER_EXITING) ? "EXITING" : "ACTIVE");
-	}
-	else if (wu.type == BACKEND_REFRESH) {
+	} else if (wu.type == WORKER_GEN && wu.payload.gen == worker_gen) {
+		return;
+	} else if (wu.type == BACKEND_REFRESH) {
 		struct backend *b;
 		b = backend_create((struct sockaddr *)&wu.payload.addr);
 		backend_deref(&backaddr);
@@ -2648,8 +2725,8 @@ handle_clear_accept(struct ev_loop *loop, ev_io *w, int revents)
 	}
 
 	CAST_OBJ_NOTNULL(fr, w->data, FRONTEND_MAGIC);
-	if (fr->ssl_ctxs != NULL)
-		CAST_OBJ_NOTNULL(so, fr->ssl_ctxs, SSLCTX_MAGIC);
+	if (fr->default_ctx != NULL)
+		CAST_OBJ_NOTNULL(so, fr->default_ctx, SSLCTX_MAGIC);
 	else
 		CAST_OBJ_NOTNULL(so, default_ctx, SSLCTX_MAGIC);
 	SSL *ssl = SSL_new(so->ctx);
@@ -2871,11 +2948,40 @@ drop_privileges(void)
 }
 
 static int
-get_backend_addrinfo(void) {
+backaddr_init_uds(void)
+{
+	struct sockaddr_un sun;
+	struct backend *b;
+	int l;
+
+	if (backaddr != NULL)
+		/* Already configured - we don't refresh UDS addresses. */
+		return (0);
+
+	AN(CONFIG->BACK_PATH);
+	memset(&sun, 0, sizeof sun);
+	sun.sun_family = PF_UNIX;
+	l = snprintf(sun.sun_path, sizeof(sun.sun_path), "%s",
+	    CONFIG->BACK_PATH);
+	/* the length of BACK_PATH is verified to fit into
+	 * sun.sun_path in configuration.c */
+	assert(l < (int)sizeof(sun.sun_path));
+
+	b = backend_create((struct sockaddr *)&sun);
+	backaddr = b;
+
+	return (0);
+}
+
+static int
+backaddr_init(void) {
 	struct addrinfo *result;
 	struct addrinfo hints;
 	struct backend *b;
 
+	if (CONFIG->BACK_PATH)
+		return (backaddr_init_uds());
+
 	memset(&hints, 0, sizeof hints);
 	hints.ai_family = AF_UNSPEC;
 	hints.ai_socktype = SOCK_STREAM;
@@ -2915,7 +3021,7 @@ init_globals(void)
 	VTAILQ_INIT(&frontends);
 	VTAILQ_INIT(&worker_procs);
 
-	get_backend_addrinfo();
+	backaddr_init();
 
 	(void)hints;
 
@@ -3075,8 +3181,13 @@ do_wait(void)
 	}
 
 	/* also check if the ocsp worker killed itself */
-	if (CONFIG->OCSP_DIR != NULL)
-		WAIT_PID(ocsp_proc_pid, start_ocsp_proc());
+	if (ocsp_proc_pid != 0)
+		WAIT_PID(ocsp_proc_pid,
+		    if (CONFIG->OCSP_DIR) {
+			    start_ocsp_proc();
+		    } else {
+			    ocsp_proc_pid = 0;
+		    });
 }
 
 static void
@@ -3625,12 +3736,12 @@ static void
 reconfigure(int argc, char **argv)
 {
 	hitch_config *cfg_new;
-	int rv;
 	struct cfg_tpc_obj_head cfg_objs;
 	struct cfg_tpc_obj *cto, *cto_tmp;
 	struct timeval tv;
 	double t0, t1;
 	struct worker_update wu;
+	struct frontend *fr;
 
 	LOGL("Received SIGHUP: Initiating configuration reload.\n");
 	AZ(gettimeofday(&tv, NULL));
@@ -3639,7 +3750,7 @@ reconfigure(int argc, char **argv)
 	VTAILQ_INIT(&cfg_objs);
 	cfg_new = config_new();
 	AN(cfg_new);
-	if (config_parse_cli(argc, argv, cfg_new, &rv) != 0) {
+	if (config_parse_cli(argc, argv, cfg_new) != 0) {
 		ERR("Config reload failed: %s\n", config_error_get());
 		config_destroy(cfg_new);
 		return;
@@ -3666,12 +3777,37 @@ reconfigure(int argc, char **argv)
 		}
 	}
 
+	/* Rewire default sslctx for each frontend after a reload */
+	VTAILQ_FOREACH(fr, &frontends, list) {
+		struct front_arg *fa;
+		struct cfg_cert_file *cf;
+		sslctx *sc;
+
+		if (HASH_COUNT(fr->ssl_ctxs) == 0) {
+			fr->default_ctx = NULL;
+			continue;
+		}
+
+		HASH_FIND_STR(cfg_new->LISTEN_ARGS, fr->pspec, fa);
+		AN(fa);
+		cf = fa->certs;
+		CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
+		while (cf->hh.next != NULL)
+			cf = cf->hh.next;
+		HASH_FIND_STR(fr->ssl_ctxs, cf->filename, sc);
+		CHECK_OBJ_NOTNULL(sc, SSLCTX_MAGIC);
+		fr->default_ctx = sc;
+	}
+
 	AZ(gettimeofday(&tv, NULL));
 	t1 = tv.tv_sec + 1e-6 * tv.tv_usec;
 
 	LOGL("{core} Config reloaded in %.2lf seconds. "
 	    "Starting new child processes.\n", t1 - t0);
 
+	config_destroy(CONFIG);
+	CONFIG = cfg_new;
+
 	worker_gen++;
 	start_workers(0, CONFIG->NCORES);
 
@@ -3679,16 +3815,15 @@ reconfigure(int argc, char **argv)
 	wu.payload.gen = worker_gen;
 	notify_workers(&wu);
 
-	if (CONFIG->OCSP_DIR != NULL) {
+	if (ocsp_proc_pid > 0) {
 		(void) kill(ocsp_proc_pid, SIGTERM);
 		/*
 		 * Restarting the OCSP process is taken
 		 * care of in do_wait
 		 */
+	} else if (CONFIG->OCSP_DIR != NULL && ocsp_proc_pid <= 0) {
+		start_ocsp_proc();
 	}
-
-	config_destroy(CONFIG);
-	CONFIG = cfg_new;
 }
 
 void
@@ -3705,7 +3840,7 @@ sleep_and_refresh(hitch_config *CONFIG)
 		rv = usleep(CONFIG->BACKEND_REFRESH_TIME*1000000);
 		if (rv == -1 && errno == EINTR)
 			break;
-		else if(get_backend_addrinfo()) {
+		else if(backaddr_init()) {
 			struct worker_update wu;
 			wu.type = BACKEND_REFRESH;
 			socklen_t len;
@@ -3725,14 +3860,13 @@ main(int argc, char **argv)
 {
 	// initialize configuration
 	struct front_arg *fa, *ftmp;
-	int rv;
 
 	CONFIG = config_new();
 
 	// parse command line
-	if (config_parse_cli(argc, argv, CONFIG, &rv) != 0) {
+	if (config_parse_cli(argc, argv, CONFIG) != 0) {
 		fprintf(stderr, "%s\n", config_error_get());
-		return (rv);
+		return (1);
 	}
 
 	if (CONFIG->TEST) {
@@ -3762,7 +3896,7 @@ main(int argc, char **argv)
 		AZ(fstat(fileno(logfile), &logf_st));
 		logf_check_t = time(NULL);
 	} else {
-		logfile = CONFIG->QUIET ? stderr : stdout;
+		logfile = stderr;
 	}
 	AZ(setvbuf(logfile, NULL, _IONBF, BUFSIZ));
 
